﻿define([
"epi",
"epi/dependency",
"dojo",
"dojo/date",
"dojo/date/stamp",
"dojo/date/locale",
"dojo/_base/lang",
"dojo/i18n"],

function (epi, dependency, dojo, dojoDate, dateStamp, locale, lang, i18n) {

    function _friendlyTimeDiff(fromDate, toDate) {
        var dayStr = epi.resources.text.day;
        var daysStr = epi.resources.text.days;

        var hourStr = epi.resources.text.hour;
        var hoursStr = epi.resources.text.hours;

        var minuteStr = epi.resources.text.minute;
        var minutesStr = epi.resources.text.minutes;

        var secondStr = epi.resources.text.second;
        var secondsStr = epi.resources.text.seconds;

        var result = "";

        var totalDiff = toDate - fromDate;

        // If the date is in the future the totalDiff will be negative, return 1 second if that is the case
        if (totalDiff < 0) {
            return " 1 " + secondStr;
        }

        var days = Math.floor(totalDiff / 1000 / 60 / 60 / 24);
        var dateDiff = new Date(totalDiff);
        var hours = dateDiff.getUTCHours();
        var minutes = dateDiff.getMinutes();
        var seconds = dateDiff.getSeconds();

        if (days !== 0) {
            result = result + " " + days + " " + (days > 1 ? daysStr : dayStr);
        }

        if (hours !== 0) {
            result = result + " " + hours + " " + (hours > 1 ? hoursStr : hourStr);
        }

        if (minutes !== 0) {
            result = result + " " + minutes + " " + (minutes > 1 ? minutesStr : minuteStr);
        }

        if (result.length > 0) {
            return result;
        }

        return " " + seconds + " " + (seconds > 1 ? secondsStr : secondStr);
    }

    function _hasProperties(/*Object*/obj) {
        // summary:
        //    Helper function that determine if the object is a simple or complex object
        //
        // obj:
        //    Object to be tested.
        //
        // tags:
        //    internal
        if (dojo.isObject(obj)) {
            for (var item in obj) {
                return true;
            }
        }
        return false;
    }

    function _transformDate(/*Object*/obj, /*boolean*/forceSimpleProperty) {
        // summary:
        //    Search property that are instance of Date
        //    in a object recursively.
        //
        // obj:
        //    The object the contain the data to be 
        //    transformed.
        //
        // tags:
        //    internal
        if (forceSimpleProperty) {
            if (obj instanceof Date) {
                return dateStamp.toISOString(obj, { zulu: true });
            }
            else {
                return obj;
            }
        }

        var newObject = dojo.clone(obj);

        for (var item in obj) {
            if (obj[item] instanceof Date) {
                newObject[item] = dateStamp.toISOString(obj[item], { zulu: true });
            }
            else {
                if (_hasProperties(obj[item])) {
                    newObject[item] = _transformDate(obj[item]);
                }
            }
        }
        return newObject;
    }

    return epi.datetime = {
        // tags:
        //      public

        // serverTimeDelta: Integer
        //      Difference time between client and server
        serverTimeDelta: 0,

        deserialize: function (str) {
            // summary:
            //    Deserialize a javascript date object from ISO string.
            //
            // date: Date
            //    The date.
            //
            // tags:
            //    public

            var rgex = /((?:\d{4})-(?:0[1-9]|1[012])-(?:0[1-9]|[12][0-9]|3[01])T(?:[01][0-9]|2[0123]):(?:[0-5][0-9])(?::(?:[0-5][0-9])(?:.\d+)?)?(?:(?:[+-](?:\d{2}):(?:[0-5][0-9]))|Z))/gi;
            if (!str) {
                return null;
            }
            else if (str.match && str.match(rgex)) {
                var date = dateStamp.fromISOString(str.match(rgex)[0]);
                return date;
            }
            else {
                throw 'Invalid JSON date';
            }
        },

        serialize: function (date) {
            // summary:
            //    Serialize a javascript date object to ISO string.
            //
            // date: Date
            //    The date to serialize.
            //
            // tags:
            //    public
            return dateStamp.toISOString(date);
        },

        transformDate: function (/*Object*/value) {
            // summary:
            //    Remove the padding server offset from the date and 
            //    convert it to ISO String.
            //
            // value:
            //    The object to be converted.
            //
            // tags:
            //    public
            var isSimpleProperty = (_hasProperties(value) && !(value instanceof Date));
            var transformedValue = (isSimpleProperty ? _transformDate(value) : _transformDate(value, true));
            return transformedValue;
        },

        isDateValid: function (/*Object*/obj) {
            // summary:
            //    Check if the date object is valid.
            //
            // obj:
            //    Object to the tested.
            //
            // tags:
            //    private
            return ((obj && obj instanceof Date) && (!isNaN(obj.getTime())));
        },

        toUserFriendlyString: function (/*Object*/date, localeName, /*boolean*/ hideToday) {
            // summary:
            //    Converts either a Date object or a date in string format to a user friendly string, for example: Today 1:51 PM
            //
            // date:
            //    The date to be displayed.
            // localeName:
            //    The locale
            // hideToday:
            //    indicates if "Today" should be removed
            // tags:
            //    public

            if (!date) {
                return "";
            }

            if (!(date instanceof Date)) {
                date = this.deserialize(date);
            }

            //Get the gregorian settings for the current locale that specifies some specific locale formats.
            var bundle = i18n.getLocalization("dojo.cldr", "gregorian");

            // Store comparison date in a new variable since we need the original still.
            var today = new Date();
            var date1 = new Date(date);

            today.setHours(0, 0, 0, 0);
            date1.setHours(0, 0, 0, 0);

            var difference = dojoDate.difference(today, date1, "day");

            // Comparisons for today (0), tomorrow (1), and yesterday (-1).
            if (difference === 0) {
                return hideToday? locale.format(date, { selector: "time", locale: localeName }) :
                                bundle["field-day-relative+0"] + " " + locale.format(date, { selector: "time", locale: localeName });
            }
            if (difference === 1) {
                return bundle["field-day-relative+1"] + " " + locale.format(date, { selector: "time", locale: localeName });
            }
            if (difference === -1) {
                return bundle["field-day-relative+-1"] + " " + locale.format(date, { selector: "time", locale: localeName });
            }

            // If date is this year format with only month and day.
            if (dojoDate.difference(today, date, "year") === 0) {
                var datePattern = bundle["dateFormatItem-MMMd"];
                return locale.format(date, { datePattern: datePattern, locale: localeName });
            }

            // Otherwise format with complete date and time.
            return locale.format(date, { locale: localeName });
        },

        toUserFriendlyHtml: function (/*Object*/date, localeName, /*boolean*/hideToday) {
            // summary:
            //    Converts either a Date object or a date in string format to a user friendly string contained in a span,
            //    for example: <span class='epi-timestamp'>Today 1:51 PM</span>.
            //
            // date:
            //    The date to be displayed.
            // localeName:
            //    The locale
            // hideToday:
            //    indicates if "Today" should be removed
            // tags:
            //    public

            var formattedDate = this.toUserFriendlyString(date, localeName, hideToday);
            if (formattedDate === "") {
                return formattedDate;
            }
            return "<span class='epi-timestamp'>" + formattedDate + "</span>";
        },

        timePassed: function (from, to) {
            // summary:
            //      Get the time passed between two dates
            //
            // from: Date
            //      Start date
            // to: Null|Date
            //      End date
            // tags:
            //      public

            to = to || new Date();
            return _friendlyTimeDiff(from, to);
        },

        timeToGo: function (/*Date*/date) {
            // summary:
            //      Get the time to go to the given time
            //
            // tags:
            //      public

            return _friendlyTimeDiff(this.serverTime(), date);
        },

        serverTime: function () {
            // summary:
            //      Get the current datetime from server
            //
            // tags:
            //      public

            return new Date(new Date().getTime() + this.serverTimeDelta);
        }
    };
});