﻿define([
"dojo/_base/array",
"dojo/_base/declare",
"dojo/_base/lang",
"dojo/store/Memory",

"epi/string",
"epi/shell/TypeDescriptorManager"],

function (array, declare, lang, MemoryStore, epiString, TypeDescriptorManager) {

    return declare(null, {
        // tags:
        //      internal

        defaultGroupType: "epi.shell.layout.GroupContainer",
        defaultSingleGroupType: "epi.shell.layout.SimpleContainer",
        defaultParentType: "epi.shell.layout.ParentContainer",
        defaultGroupName: "defaultGroup",
        defaultGroupTitle: "",

        propertyFilter: null,
        _item: null,

        constructor: function (args) {
            lang.mixin(this, args);
        },

        toComponentDefinitions: function (item /*object*/, nameBase /*string*/, useDefaultValue /*Boolean*/, readOnly) {
            // summary:
            //     transform the data into a format that epi.shell.widget.WidgetFactory can process.

            // item: Array
            //     The original item in the metadata sent by the server

            // nameBase: String
            //     The name of the level, empty to start with,
            //    get the parents property names concatenated during the recursivity if there are children

            // useDefaultValue: Boolean
            //     Flag indicating if the property will get its value or the default value.

            // readOnly: Boolean
            //  Flag indicating if the property is read only

            this._item = item;
            var properties = this._filterProperties(item);

            var componentDefinitions = this._getComponentDefinitions(properties, item.groups, nameBase, useDefaultValue, readOnly);

            return componentDefinitions;
        },

        _filterProperties: function (property) {
            var filter = lang.isFunction(this.propertyFilter) ? lang.hitch(this, function (prop) { return this.propertyFilter(property, prop); }) : function () { return true; },
            properties = array.filter(property.properties, filter);
            return properties;
        },

        _getComponentDefinitions: function (propertyDefinitions, groupDefinitions, nameBase, useDefaultValue, readOnly) {
            var groupContainers = [],
                groups = lang.clone(groupDefinitions) || [],
                properties = lang.clone(propertyDefinitions) || [],

                sortFn = function (item1, item2) {
                    return (item1.displayOrder || 0) - (item2.displayOrder || 0);
                };

            // Sort groups and properties based on displayOrder
            groups.sort(sortFn);
            properties.sort(sortFn);

            // Create group containers
            array.forEach(groups, lang.hitch(this, function (group) {
                var groupContainer = this._createGroupContainer(group, readOnly);
                groupContainers.push(groupContainer);
            }));

            // Loop through property definitions, create property components, and add to groups
            array.forEach(properties, lang.hitch(this, function (property) {
                if (!property.showForEdit) {
                    return;
                }

                // Create property component
                var propertyComponent = property.properties && property.properties.length ?
                    this._createParentContainer(property, nameBase, useDefaultValue, readOnly) : // Create sub structure if have sub properties
                    this._createPropertyWidget(property, nameBase, useDefaultValue, readOnly);   // Otherwise, create property widget

                // Find a group component to add
                var groupFound = array.some(groupContainers, lang.hitch(this, function (groupContainer) {
                    if (groupContainer.settings.name === epiString.pascalToCamel(property.groupName || this.defaultGroupName)) {
                        groupContainer.components.push(propertyComponent);
                        return true;
                    }
                    return false;
                }));

                if (!groupFound) {

                    // Not found, create a default group container and add to the top
                    var defaultGroup = this._createGroupContainer({
                        name: this.defaultGroupName,
                        title: this.defaultGroupTitle,
                        displayUI: true
                    }, readOnly);

                    defaultGroup.components.push(propertyComponent);
                    groupContainers = [defaultGroup].concat(groupContainers);
                }
            }));

            // Return group containers that will be visible and contains at least 1 component.
            var filteredGroupContainers = array.filter(groupContainers, function (item) {
                return item.settings.displayui && item.components && item.components.length;
            });

            // Set default group type
            var defaultGroupType = (filteredGroupContainers.length > 1) ? this.defaultGroupType : this.defaultSingleGroupType;
            array.forEach(filteredGroupContainers, lang.hitch(this, function (item) {
                item.widgetType = item.widgetType || defaultGroupType;
            }));

            return filteredGroupContainers;
        },

        _createGroupContainer: function (group, readOnly) {
            var baseSettings = {
                _type: "group",
                name: epiString.pascalToCamel(group.name),
                title: group.title,
                readOnly: readOnly,
                displayui: group.displayUI
            };

            return {
                settings: lang.mixin(baseSettings, group.options),
                widgetType: group.uiType,
                components: []
            };
        },

        _createParentContainer: function (property, nameBase, useDefaultValue, readOnly) {
            var baseSettings = {
                _type: "parent",
                name: nameBase + epiString.pascalToCamel(property.name),
                title: property.displayName
            };

            var properties = this._filterProperties(property);

            return {
                settings: lang.mixin(baseSettings, property.settings),
                widgetType: property.layoutType || this.defaultParentType,
                components: this._getComponentDefinitions(properties, property.groups, baseSettings.name + '.', useDefaultValue, readOnly)
            };
        },

        _createPropertyWidget: function (property, nameBase, useDefaultValue, readOnly) {
            var baseSettings = lang.mixin(useDefaultValue ? { value: property.initialValue } : {}, {
                _type: property.hideSurroundingHtml ? "hiddenfield" : "field",
                name: nameBase + epiString.pascalToCamel(property.name),
                label: property.displayName,
                selections: property.selections,
                displayui: property.displayui,
                readOnly: readOnly
            }, property.settings);

            lang.mixin(baseSettings, this._getTranslatedLabelAndTooltip(property.name));

            return {
                settings: baseSettings,
                widgetType: property.uiType,
                widgetPackage: property.uiPackage
            };
        },

        _getTranslatedLabelAndTooltip: function (propertyName) {
            var returnValue = {};

            if (!this._item.additionalValues || !this._item.additionalValues.modelTypeIdentifier) {
                return returnValue;
            }

            var property = TypeDescriptorManager.getResourceValue(this._item.additionalValues.modelTypeIdentifier, "properties." + propertyName.toLowerCase());
            if (!property) {
                return returnValue;
            }

            if (property.caption) {
                returnValue.label = property.caption;
            }

            if (property.help) {
                returnValue.tooltip = property.help;
            }

            return returnValue;
        }
    });
});

