﻿define([
// dojo
    "dojo",

    "dojo/_base/array",
    "dojo/_base/declare",
    "dojo/_base/lang",
    "dojo/when",

    "epi/dependency",

// epi
    "epi/i18n!epi/shell/ui/nls/contenttypes"
],

function (
// dojo
    dojo,

    array,
    declare,
    lang,
    when,

    dependency,

// epi
    typeResources
) {

    return {
        // summary:
        //    This class manages ui behavior for classes including resolving inherited values.
        // tags:
        //    public

        // _uiDescriptors: Array
        //    An array of descriptors
        _uiDescriptors: {},

        resources: typeResources,

        store: null,

        initialize: function () {
            this.store = this.store || (function () {
                var registry = dependency.resolve("epi.storeregistry");
                return registry.get("epi.shell.uidescriptor");
            })();

            return when(this.store.query(), lang.hitch(this, function (uiDescriptors) {
                array.forEach(uiDescriptors, function (uiDescriptor) {
                    this._uiDescriptors[uiDescriptor.typeIdentifier] = uiDescriptor;
                }, this);
            }));
        },

        clear: function () {
            // summary:
            //    Clears all registered settings.
            this._uiDescriptors = {};
        },

        registerTypeSetting: function (name, setting) {
            // summary:
            //    Gets a given setting for a type

            this._uiDescriptors[name] = setting;
        },

        getValue: function (type, settingName) {
            // summary:
            //    Gets a given setting for a type
            //
            // type: string
            //    The type that you want to fetch the settings for.
            //
            // settingName: string
            //    The settings that you want to fetch.

            var typeDescriptor = this._uiDescriptors[type];
            while (typeDescriptor) {
                var setting = typeDescriptor[settingName];

                if (setting !== null && setting !== undefined) {
                    return setting;
                }

                typeDescriptor = this._uiDescriptors[typeDescriptor["baseTypeIdentifier"]];
            }
        },

        getResourceValue: function (type, resourceName, primaryType) {
            // summary:
            //      Gets a resource for a given type.
            // type: [String]
            //      The type that you want to fetch the settings for.
            // resourceName: [String]
            //      The settings that you want to fetch.
            // primaryType: [Boolean]
            //      Flag indicates that needed to get resources from primary type or not
            // tags:
            //      public

            var typeDescriptor = this._uiDescriptors[type];
            while (typeDescriptor) {
                var typeResources = this.resources[typeDescriptor.languageKey];
                if (typeResources && (primaryType ? typeDescriptor.isPrimaryType === true : true)) {
                    //Handle nested resources
                    var resourceObject = lang.getObject(resourceName, false, typeResources);
                    if (resourceObject) {
                        return resourceObject;
                    }
                }

                typeDescriptor = this._uiDescriptors[typeDescriptor["baseTypeIdentifier"]];
            }
        },

        getAndConcatenateValues: function (type, settingName) {
            // summary:
            //    Gets a given setting for a type and it's ancestors returning an array with all values.
            //
            // type: string
            //    The type that you want to fetch the settings for.
            //
            // settingName: string
            //    The settings that you want to fetch.

            var concatenatedValues = [],
                    typeDescriptor = this._uiDescriptors[type];
            while (typeDescriptor) {
                var setting = typeDescriptor[settingName];
                if (setting) {
                    if (lang.isArray(setting)) {
                        array.forEach(setting, function (item) {
                            concatenatedValues.push(item);
                        });
                    }
                    else {
                        concatenatedValues.push(setting);
                    }
                }
                typeDescriptor = this._uiDescriptors[typeDescriptor["baseTypeIdentifier"]];
            }

            return concatenatedValues;
        },

        getInheritanceChain: function (type) {
            // summary:
            //    Returns an array containing the type identifiers for the given type and it's inherited types.
            //
            // type: string
            //    The type that you want to fetch the inheritance chain for.

            var concatenatedValues = [],
                    typeDescriptor = this._uiDescriptors[type];
            while (typeDescriptor) {
                concatenatedValues.push(typeDescriptor.typeIdentifier);
                typeDescriptor = this._uiDescriptors[typeDescriptor["baseTypeIdentifier"]];
            }

            return concatenatedValues;
        },

        isBaseTypeIdentifier: function (childType, parentType) {
            // summary: 
            //      Test if parent type is base of child type/
            // childType: String
            //      the child type
            // parentType: String
            //      the parent type

            var type = childType,
                descriptor;

            if (childType === parentType) {
                return true;
            }

            while (descriptor = this._uiDescriptors[type]) {
                type = descriptor.baseTypeIdentifier;

                if (type === parentType) {
                    return true;
                }
            }

            return false;
        }
    };
});