﻿define([
    "dojo",
    "epi"
],
    function (dojo, epi) {

        return dojo.declare(null, {
            // summary:
            //		Class that helps positioning elements in in the vicinity of other elements, 
            //      taking frames, scroll, window size, and element (re)sizes into consideration.
            //
            // example:
            //	|	var poser = new epi.shell.PositioningUtility();
            //	|	poser.place(widget, { placement:'nextTo', elementInFrame: node });
            //
            // example:
            //	|	var poser = new epi.shell.PositioningUtility();
            //	|	poser.place(widget, { placement:'over', position: { x: 100, y:200 }, displayCallback:widget.loaded, resizeCallback:widget.resize });
            //
            // tags: 
            //      private

            _margin: 10,

            constructor: function () { },

            place: function (/*Object*/elementToPosition, /*Object*/settings) {
                // summary:
                //      Positions elementToPosition close but not over settings.snapTo 
                //      using the lightweight floating window ruleset.
                // elementToPosition:
                //      The node|id|position we want to get a position for.
                // settings:
                //      An object containing settings:
                //          - snapTo: The node|id|position elementToPosition should dock to.

                var node = this._getNode(elementToPosition);

                var pos = this.findPosition(node, settings);

                dojo.style(node, { position: "absolute", top: pos.y + "px", left: pos.x + "px" });
            },

            findPosition: function (/*Object*/elementToPosition, /*Object*/settings) {
                // summary:
                //      Gets the best suitable position for elementToPosition close but 
                //      not over settings.snapTo using the lightweight floating window ruleset.
                // elementToPosition:
                //      The node|id|position we want to get a position for.
                // settings:
                //      An object containing settings:
                //          - snapTo: The node|id|position elementToPosition should dock to.

                var dialogNode = this._getNode(elementToPosition);
                var dialog = this._getPosition(dialogNode);
                var dialogViewport = dojo.window.getBox();

                if (settings.padding) {
                    dialog.w += settings.padding.w;
                    dialog.h += settings.padding.h;
                }

                //console.log("dialog position ", dialog);

                var blockNode = this._getNode(settings.snapTo);
                var block = this._getPosition(settings.snapTo);
                var blockViewport = settings.frame || { x: 0, y: 0, w: dialogViewport.w, h: dialogViewport.h };

                if (settings.frame == null && dialogNode != null && blockNode != null) {
                    var blockNodeWindow = 'defaultView' in blockNode.ownerDocument ? blockNode.ownerDocument.defaultView : blockNode.ownerDocument.parentWindow;
                    var dialogNodeWindow = 'defaultView' in dialogNode.ownerDocument ? dialogNode.ownerDocument.defaultView : dialogNode.ownerDocument.parentWindow;
                    blockViewport = this._getFramePosition(blockNodeWindow, dialogNodeWindow);
                }

                //console.log("block position ", block);

                block = this._getVisiblePosition(block, blockViewport);

                //console.log("block position in top window ", block);

                return this._executeTransformers(dialog, block, dialogViewport)
            || this._fallbackTransformer(dialog, block, dialogViewport);
            },

            _getVisiblePosition: function (block, blockViewport) {
                // get the visible portion of the block and offset the position to the outer frame's coordinates
                var visibleBlock = dojo.clone(block);
                if (blockViewport.w > 0 && visibleBlock.x + visibleBlock.w > blockViewport.w)
                    visibleBlock.w = blockViewport.w - visibleBlock.x;
                if (blockViewport.h > 0 && visibleBlock.y + visibleBlock.h > blockViewport.h)
                    visibleBlock.h = blockViewport.h - visibleBlock.y;
                visibleBlock.x += blockViewport.x;
                visibleBlock.y += blockViewport.y;

                if (visibleBlock.y < blockViewport.y) {
                    visibleBlock.h = visibleBlock.h - blockViewport.y + visibleBlock.y;
                    visibleBlock.y = blockViewport.y;
                }

                if (visibleBlock.x < blockViewport.x) {
                    visibleBlock.w = visibleBlock.w - blockViewport.x + visibleBlock.x;
                    visibleBlock.x = blockViewport.x;
                }

                return visibleBlock;
            },

            _getFramePosition: function (blockWindow, dialogWindow) {
                if (blockWindow.parent !== blockWindow) {
                    var iframes = blockWindow.parent.document.getElementsByTagName("iframe");
                    for (var i in iframes) {
                        //console.log("considering frame ", iframes[i], iframes[i].src);
                        if (iframes[i].contentWindow === blockWindow) {
                            var positionWithinParentWindow = dojo.position(iframes[i]);
                            if (blockWindow.parent === dialogWindow)
                                return positionWithinParentWindow;

                            var positionWithinAncestorWindow = this._getFramePosition(blockWindow.parent, dialogWindow);

                            //console.log("parent/ancestor ", positionWithinParentWindow, positionWithinAncestorWindow);

                            return {
                                x: positionWithinParentWindow.x + positionWithinAncestorWindow.x,
                                y: positionWithinParentWindow.y + positionWithinAncestorWindow.y,
                                w: positionWithinParentWindow.w,
                                h: positionWithinParentWindow.h
                            };
                        }
                    }
                }

                return { x: 0, y: 0, w: 0, h: 0 };
            },

            _executeTransformers: function (pos, block, viewport) {
                for (var i in this._transformers) {
                    var posToTest = this._transformers[i].call(this, pos, block, viewport);
                    posToTest = this._moveIntoViewport(posToTest, viewport);
                    if (this._isInside(posToTest, viewport) && this._isOutside(posToTest, block)) {
                        //console.log("placing at ", posToTest);
                        return posToTest;
                    }
                    //else
                    //    console.log("considered position ", posToTest);
                }
                return null;
            },

            _getNode: function (idOrWidgetOrNode) {
                if (typeof idOrWidgetOrNode === "string")
                    return dojo.byId(idOrWidgetOrNode);

                var node = idOrWidgetOrNode;
                if (node && node.domNode)
                    return node.domNode;

                return node;
            },

            _getPosition: function (idOrWidgetOrNodeOrPosition) {
                if (typeof idOrWidgetOrNodeOrPosition == "undefined" || idOrWidgetOrNodeOrPosition === null)
                    return { x: 0, y: 0, w: 0, h: 0 };

                var pos = idOrWidgetOrNodeOrPosition;
                if (typeof pos.x !== "undefined" && typeof pos.y !== "undefined")
                    return pos;

                var node = this._getNode(idOrWidgetOrNodeOrPosition);
                if (!node)
                    return { x: 0, y: 0, w: 0, h: 0 };

                return dojo.position(node, false);
            },

            _moveIntoViewport: function (dialog, viewport) {
                return {
                    x: Math.min(viewport.w - dialog.w - this._margin, Math.max(this._margin, dialog.x)),
                    y: Math.min(viewport.h - dialog.h - this._margin, Math.max(this._margin, dialog.y)),
                    w: dialog.w,
                    h: dialog.h
                };
            },

            _isInside: function (dialog, viewport) {
                return dialog.x > 0
            && dialog.y > 0
            && dialog.x + dialog.w < viewport.w
            && dialog.y + dialog.h < viewport.h;
            },

            _isOutside: function (dialog, block) {
                return dialog.x > block.x + block.w
            || dialog.y > block.y + block.h
            || dialog.x + dialog.w < block.x
            || dialog.y + dialog.h < block.y;
            },

            _transformers: [
            function (dialog, block, viewport) {
                //pos 1 = right align top
                return {
                    x: block.x + block.w + this._margin,
                    y: block.y,
                    w: dialog.w,
                    h: dialog.h
                };
            },
            function (dialog, block, viewport) {
                //pos 2 = centered below
                return {
                    x: block.x + (block.w / 2 - dialog.w / 2),
                    y: block.y + block.h + this._margin,
                    w: dialog.w,
                    h: dialog.h
                };
            },
            function (dialog, block, viewport) {
                //pos 3 = left align top
                return {
                    x: block.x - dialog.w - this._margin,
                    y: block.y,
                    w: dialog.w,
                    h: dialog.h
                };
            },
            function (dialog, block, viewport) {
                //pos 4 = centered above
                return {
                    x: block.x + (block.w / 2 - dialog.w / 2),
                    y: block.y - dialog.h - this._margin,
                    w: dialog.w,
                    h: dialog.h
                };
            }
        ],
            _fallbackTransformer: function (dialog, block, viewport) {
                return {
                    x: (viewport.w - dialog.w - this._margin),
                    y: 10
                };
            }
        });
    });
