﻿define([
    "dojo/_base/declare",
    "dojo/_base/lang",

    "dojo/dom-style",

    "dijit/_WidgetsInTemplateMixin",

    "dijit/focus",
    "dijit/Dialog",
    "dijit/DialogUnderlay",

    "epi",
    "epi/shell/widget/dialog/_DialogMixin",

    "dojo/text!./templates/LightWeight.html"
],
function (
    declare,
    lang,

    domStyle,

    _WidgetsInTemplateMixin,

    focus,
    Dialog,
    DialogUnderlay,

    epi,
    _DialogMixin,

    template
) {

    return declare([Dialog, _DialogMixin, _WidgetsInTemplateMixin], {
        // summary:
        //		A lightweight dialog Widget
        //
        // description:
        //		Pops up a modeless lightweight dialog window
        //
        // tags:
        //      public

        // templateString: [protected] String
        //		A string that represents the widget template.
        templateString: template,

        // buttonCloseText: [protected] String
        //      Text to be displayed on the 'close' button
        buttonCloseText: epi.resources.action.close,

        // buttonDoneText: [protected] String
        //      Text to be displayed on the 'done' button
        buttonDoneText: epi.resources.action.done,

        // buttonLearnMoreText: [protected] String
        //      Text to be displayed on the 'learnmore' button
        buttonLearnMoreText: epi.resources.action.learnmore,

        // useGlassUnderlay: [public] Boolean
        //    Denotes that a glass underlay should be used. This underlay is not the global dijit._underlay which is designed for modal dialogs only.
        useGlassUnderlay: false,

        // _glassUnderlay: [public] dijit.DialogUnderlay
        //    Underlay object. dijit.Dialog use a single global underlay which is not fully transparent for all dialog layers.
        //    This lightweight dialog maintains a different underlay to avoid messing up dijit's dialog stack.
        _glassUnderlay: null,

        "class": "epi-lfw-dialog",

        // snapToNode: [public] domNode
        //    The DomNode to snap to when doing positioning.
        snapToNode: null,

        // positioner: [public] epi.shell.layout.PositioningUtility
        //    Utility used to determine the correct position to display this dialog.
        positioner: null,

        destroy: function () {
            // summary:
            //		Destroy dialog

            //delete used underlay
            if (this._glassUnderlay) {
                this._glassUnderlay.destroyRecursive();
                delete this._glassUnderlay;
            }

            this.inherited(arguments);
        },

        show: function () {
            // summary:
            //		Display the dialog
            // returns: dojo.Deferred
            //		Deferred object that resolves when the display animation is complete

            return this.inherited(arguments).then(lang.hitch(this, this._onAfterShow));
        },

        _size: function () {
            this.inherited(arguments);

            // Optional property. Check null before using
            if (this.positioner) {
                this._relativePosition = this.positioner.findPosition(this.domNode, { snapTo: this.snapToNode });
            }
        },

        resize: function () {
            this._size();
            this._position();
        },

        _onAfterShow: function () {
            // summary:
            //		Additional work after dialog is shown.

            if (this.content && this.content.focus) {
                this.content.focus();
            }

            if (this.useGlassUnderlay) {
                this._createUnderlay();
            }
            else {
                //No glass underlay created when showing up, we use a temporary underlay to avoid lag when moving/resizing.
                //Handle moving
                if (this._moveable && !this._attachedMoveEvents) {
                    this.connect(this._moveable, 'onMoveStart', lang.hitch(this, this._createUnderlay));
                    this.connect(this._moveable, 'onMoveStop', lang.hitch(this, this._removeUnderlay));
                    this._attachedMoveEvents = true;
                }

                //Handle content resizing
                if (this.content && this.content.onResizeStart && this.content.onResizeStop && !this._attachedContentResizeEvents) {
                    this.connect(this.content, 'onResizeStart', lang.hitch(this, this._createUnderlay));
                    this.connect(this.content, 'onResizeStop', lang.hitch(this, this._removeUnderlay));
                    this._attachedContentResizeEvents = true;
                }
            }

            this.onAfterShow();
        },

        _createUnderlay: function () {
            // summary:
            //		Handle move start event.
            // description:
            //      Show a transparent underlay to prevent lagging when drag over the underneath iframe.

            if (this._glassUnderlay) {
                this._glassUnderlay.show();
            }
            else {
                this._glassUnderlay = new DialogUnderlay();
                this._glassUnderlay.show();

                // Set z-index a bit under the dialog
                var zIndex = domStyle.get(this.domNode, 'zIndex');
                domStyle.set(this._glassUnderlay.domNode, 'zIndex', zIndex - 1);

                //Make it transparent
                domStyle.set(this._glassUnderlay.domNode, 'filter', 'alpha(opacity=0)');
                domStyle.set(this._glassUnderlay.domNode, 'opacity', '0');
            }
        },

        _removeUnderlay: function () {
            // summary:
            //		Handle move stop event.
            // description:
            //      Hide the transparent underlay.

            this._glassUnderlay.hide();
        },

        _onContainerNodeClick: function (e) {
            //If we inside the dialog, set the focus to the first focusable item
            if (e.target === this.containerNode) {
                this._getFocusItems();
                this._firstFocusItem && focus.focus(this._firstFocusItem);
            }
        },

        onButtonClose: function () {

        },

        onAfterShow: function () {
            // summary:
            //    Fired on after show.
            //
            // tags:
            //    public, callback
        },

        onAfterHide: function () {
            // summary:
            //    Fired on hide.
            //
            // tags:
            //    public, callback
        },

        hide: function () {
            // summary:
            //		Hide the dialog
            // returns: dojo.Deferred
            //		Deferred object that resolves when the hide animation is complete

            return this.inherited(arguments).then(lang.hitch(this, this._onAfterHide));
        },

        _onAfterHide: function () {
            if (this.useGlassUnderlay) {
                this._removeUnderlay();
            }

            this.onAfterHide();
        },

        layout: function () {
            // summary:
            //		Position the Dialog and the underlay
            // tags:
            //		private

            if (this._glassUnderlay && domStyle.get(this._glassUnderlay.domNode, "display") != "none") {
                // Calling layout of the underlay sets display, and since show() calls layout we can ignore it if it's not shown
                this._glassUnderlay.layout();
            }

            var underlay = DialogUnderlay._singleton;

            //Adjust other dijit underlay display state, since Dialog.layout always set the modal Dialog's underlay to visible.
            var dijitUnderlayDisplay = underlay ? domStyle.get(underlay.domNode, "display") : null;

            this.inherited(arguments);

            if (underlay) {
                domStyle.set(underlay.domNode, "display", dijitUnderlayDisplay);
            }
        }
    });
});
